import numpy as np
import numpy.testing as npt

import argosim.imaging_utils as aiu


class TestImagingUtils:

    sky_model_params = (
        (256, 256),  # shape_px
        1.0,  # fov
        [0.01, 0.02, 0.03],  # deg_size_list
        [0.4, 0.3, 0.3],  # source_intensity_list
        332,  # seed
    )
    sky_model_expected_path = "src/argosim/tests/data/sky_model_exp.npy"
    sky_model_uv_expected_path = "src/argosim/tests/data/sky_model_uv_exp.npy"

    pathfinder_uv_track_path = "src/argosim/tests/data/pathfinder_uv_track.npy"
    grid_uv_samples_params = (
        (256, 256),  # Image shape in pixels
        (3.0, 3.0),  # FOV in degrees
    )
    pathfinder_uv_mask_path = (
        "src/argosim/tests/data/argos_pathfinder_uv_mask_binary.npy"
    )
    pathfinder_uv_mask_hist_path = (
        "src/argosim/tests/data/argos_pathfinder_uv_mask_hist.npy"
    )
    pathfinder_uv_mask_weighted_path = (
        "src/argosim/tests/data/argos_pathfinder_uv_mask_weighted.npy"
    )
    uv_weights_path = "src/argosim/tests/data/uv_sampling_weights.npy"

    sky_uv_w_masked_path = "src/argosim/tests/data/sky_uv_w_masked.npy"

    uv_noise_params = (0.1, 612)  # noise_level  # seed
    sky_uv_w_masked_noisy_path = "src/argosim/tests/data/sky_uv_w_masked_noisy.npy"

    obs_sim_single_band_path = "src/argosim/tests/data/obs_sim_single_band.npy"
    dirty_beam_sim_single_band_path = (
        "src/argosim/tests/data/dirty_beam_sim_single_band.npy"
    )
    obs_sim_single_band_params = {
        "fov_size": 1.0,
        "sigma": 0.1,
        "seed": 717,
        "freqs": [1.5e9],
    }
    decimal_uv = 4

    def test_sky2uv(self):
        sky = np.load(self.sky_model_expected_path)
        sky_uv = aiu.sky2uv(sky)
        sky_uv_expected = np.load(self.sky_model_uv_expected_path)
        npt.assert_array_almost_equal(
            sky_uv,
            sky_uv_expected,
            decimal=self.decimal_uv,
            err_msg="Sky to UV conversion failed. The resulting UV image does not match the expected output.",
        )

    def test_uv2sky(self):
        sky_uv = np.load(self.sky_model_uv_expected_path)
        sky = aiu.uv2sky(sky_uv)
        sky_expected = np.load(self.sky_model_expected_path)
        npt.assert_array_almost_equal(
            sky,
            sky_expected,
            err_msg="UV to Sky conversion failed. The resulting Sky image does not match the expected output.",
        )

    def test_grid_uv_samples(self):
        track = np.load(self.pathfinder_uv_track_path)
        # Test binary mask
        mask_uv, _ = aiu.grid_uv_samples(
            track, *self.grid_uv_samples_params, mask_type="binary"
        )
        mask_uv_expected = np.load(self.pathfinder_uv_mask_path)
        npt.assert_array_almost_equal(
            mask_uv,
            mask_uv_expected,
            err_msg="Binary mask UV samples do not match the expected output.",
        )

        # Test histogram mask
        mask_uv_hist, _ = aiu.grid_uv_samples(
            track, *self.grid_uv_samples_params, mask_type="histogram"
        )
        mask_uv_hist_expected = np.load(self.pathfinder_uv_mask_hist_path)
        npt.assert_array_almost_equal(
            mask_uv_hist,
            mask_uv_hist_expected,
            err_msg="Histogram mask UV samples do not match the expected output.",
        )

        # Test weighted mask
        weights = np.load(self.uv_weights_path)
        mask_uv_weighted, _ = aiu.grid_uv_samples(
            track, *self.grid_uv_samples_params, mask_type="weighted", weights=weights
        )
        mask_uv_weighted_expected = np.load(self.pathfinder_uv_mask_weighted_path)
        npt.assert_array_almost_equal(
            mask_uv_weighted,
            mask_uv_weighted_expected,
            err_msg="Weighted mask UV samples do not match the expected output.",
        )

    def test_grid_uv_samples_out_of_range(self):
        track = np.load(self.pathfinder_uv_track_path)
        # catch ValueError for out of range samples
        with npt.assert_raises(ValueError):
            mask_uv, _ = aiu.grid_uv_samples(track, (128, 128), (3.0, 3.0))
        with npt.assert_raises(ValueError):
            mask_uv, _ = aiu.grid_uv_samples(
                track, (64, 64), (1.0, 1.0), mask_type="histogram"
            )

    def test_grid_uv_samples_invalid_mask_type(self):
        track = np.load(self.pathfinder_uv_track_path)
        # catch ValueError for invalid mask type
        with npt.assert_raises(ValueError):
            mask_uv, _ = aiu.grid_uv_samples(
                track, *self.grid_uv_samples_params, mask_type="invalid_mask_type"
            )

    def test_grid_uv_samples_missing_weights(self):
        track = np.load(self.pathfinder_uv_track_path)
        # catch AssertionError for missing weights when mask_type is 'weighted'
        with npt.assert_raises(AssertionError):
            mask_uv, _ = aiu.grid_uv_samples(
                track, *self.grid_uv_samples_params, mask_type="weighted"
            )

    def test_compute_visibilities_grid(self):
        sky_uv = np.load(self.sky_model_uv_expected_path)
        mask_uv = np.load(self.pathfinder_uv_mask_weighted_path)
        sky_uv_w_masked_out = aiu.compute_visibilities_grid(sky_uv, mask_uv)
        sky_uv_w_masked_exp = np.load(self.sky_uv_w_masked_path)
        npt.assert_array_almost_equal(
            sky_uv_w_masked_out,
            sky_uv_w_masked_exp,
            err_msg="Computed grided visibilities do not match the expected output.",
        )

    def test_add_noise_uv(self):
        sky_uv_masked = np.load(self.sky_uv_w_masked_path)
        mask_uv = np.load(self.pathfinder_uv_mask_weighted_path)
        sky_uv_masked_noisy_out = aiu.add_noise_uv(
            sky_uv_masked, mask_uv, *self.uv_noise_params
        )
        sky_uv_masked_noisy_exp = np.load(self.sky_uv_w_masked_noisy_path)
        npt.assert_array_almost_equal(
            sky_uv_masked_noisy_out,
            sky_uv_masked_noisy_exp,
            decimal=self.decimal_uv,
            err_msg="Adding noise to UV samples did not produce the expected output.",
        )

    def test_simulate_dirty_obs_single_band(self):
        sky = np.load(self.sky_model_expected_path)
        track = np.load(self.pathfinder_uv_track_path)
        params = self.obs_sim_single_band_params
        obs_out, dirty_beam_out = aiu.simulate_dirty_observation(
            sky,
            track,
            fov_size=params["fov_size"],
            sigma=params["sigma"],
            seed=params["seed"],
        )
        obs_exp = np.load(self.obs_sim_single_band_path)
        dirty_beam_exp = np.load(self.dirty_beam_sim_single_band_path)

        npt.assert_array_almost_equal(
            obs_out,
            obs_exp,
            err_msg="Simulated dirty observation does not match the expected output.",
        )
        npt.assert_array_almost_equal(
            dirty_beam_out,
            dirty_beam_exp,
            err_msg="Simulated dirty beam does not match the expected output.",
        )

    def test_simulate_dirty_obs_multi_band(self):
        sky = np.load(self.sky_model_expected_path)
        track = np.load(self.pathfinder_uv_track_path)
        params = self.obs_sim_single_band_params
        obs_out_multi, dirty_beam_out_multi = aiu.simulate_dirty_observation(
            sky,
            np.expand_dims(track, axis=0),
            fov_size=params["fov_size"],
            sigma=params["sigma"],
            seed=params["seed"],
            multi_band=True,
            freqs=params["freqs"],
        )
        obs_exp = np.load(self.obs_sim_single_band_path)
        dirty_beam_exp = np.load(self.dirty_beam_sim_single_band_path)

        npt.assert_array_almost_equal(
            obs_out_multi[0],
            obs_exp,
            err_msg="Simulated multi-band dirty observation does not match the expected output.",
        )
        npt.assert_array_almost_equal(
            dirty_beam_out_multi[0],
            dirty_beam_exp,
            err_msg="Simulated multi-band dirty beam does not match the expected output.",
        )



"""Antenna utils.

This module contains functions to generate antenna arrays, compute its baselines,
perform aperture synthesis, obtain uv-coverage and get observations from sky models.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>
          Samuel Gullin <gullin@ia.forth.gr>

"""

import jax.numpy as jnp
import numpy as np
import numpy.random as rnd
from jax import jit, vmap

from argosim.rand_utils import local_seed

########################################
#      Generate antenna positions      #
########################################


def random_antenna_pos(E_lim=1000.0, N_lim=1000.0, U_lim=0.0, seed=None):
    """Random antenna pos.

    Function to generate a random antenna location in ENU coordinates.
    Antenna lies in the range: [-E_lims/2, E_lims/2]x[-N_lims/2, N_lims/2]x[0, U_lims].

    Parameters
    ----------
    E_lim : float
        The east coordinate span width of the antenna position in meters.
    N_lim : float
        The north coordinate span width of the antenna position in meters.
    U_lim : float
        The up coordinate span width of the antenna position in meters.
    seed : int
        Optional seed to set.

    Returns
    -------
    antenna_pos : np.ndarray
        The antenna position in ENU coordinates.
    """
    with local_seed(seed):
        random_coords = rnd.random_sample(3)

    # Return (x,y) random location for single dish
    return (
        random_coords * np.array([E_lim, N_lim, U_lim])
        - np.array([E_lim, N_lim, 0.0]) / 2
    )


def circular_antenna_arr(n_antenna=3, r=300.0):
    """Circular antenna arr.

    Function to generate a circular antenna array. Antennas lie in a circumference
    of radius 'r' from the center [0,0] and are equally spaced.

    Parameters
    ----------
    n_antenna : int
        The number of antennas in the array.
    r : float
        The radius of the antenna array in meters.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n' antenna locations (x_i, y_i) equally spaced over a 'r' radius circumference.
    return np.array(
        [
            [np.cos(angle) * r, np.sin(angle) * r, 0.0]
            for angle in [2 * np.pi / n_antenna * i for i in range(n_antenna)]
        ]
    )


def y_antenna_arr(n_antenna=5, r=500.0, alpha=0.0):
    """Y antenna arr.

    Function to generate a Y-shaped antenna array. Antennas lie equispaced in three radial arms
    of 120 degrees each.

    Parameters
    ----------
    n_antenna : int
        The number of antennas per arm.
    r : float
        The radius of the antenna array in meters.
    alpha : float
        The angle of the first arm.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n' antenna locations (x_i, y_i) equispaced on three (120 deg) radial arms.
    step = r / n_antenna
    return np.array(
        [
            [
                np.array(
                    [
                        (i + 1) * step * np.cos(angle / 180 * np.pi),
                        (i + 1) * step * np.sin(angle / 180 * np.pi),
                        0.0,
                    ]
                )
                for i in range(n_antenna)
            ]
            for angle in [alpha, alpha + 120, alpha + 240]
        ]
    ).reshape((3 * n_antenna, 3))


def random_antenna_arr(n_antenna=3, E_lim=1000.0, N_lim=1000.0, U_lim=0.0, seed=None):
    """Random antenna arr.

    Function to generate a random antenna array. Antennas lie randomly distributed
    in the range: [-E_lims/2, E_lims/2]x[-N_lims/2, N_lims/2]x[0, U_lims].

    Parameters
    ----------
    n_antenna : int
        The number of antennas in the array.
    E_lim : float
        The east coordinate span width of the antenna positions in meters.
    N_lim : float
        The north coordinate span width of the antenna positions in meters.
    U_lim : float
        The up coordinate span width of the antenna positions in meters.
    seed : int
        Optional seed to set.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    with local_seed(seed):
        # Make a list of 'n' antenna locations (x_i, y_i) randomly distributed.
        positions = [random_antenna_pos(E_lim, N_lim, U_lim) for i in range(n_antenna)]

    return np.array(positions)


def uni_antenna_array(
    n_antenna_E=32, n_antenna_N=32, E_lim=800.0, N_lim=800.0, U_lim=0.0
):
    """Uniform (grid) antenna arr.

    Function to generate a uniform antenna array. Antennas lie uniformely distributed
    in the range, and center on [0,0,U_lim]: [-E_lim/2, E_lim/2]x[-N_lim/2, N_lim/2]x[U_lim].
    Only allow to generate a grid at a fixed height U_lim (fixed Up component).

    Parameters
    ----------
    n_antenna_E : int
        The number of antennas in the North direction (vertical).
    n_antenna_N : int
        The number of antennas in the East direction (horizontal).
    E_lim : float
        The east coordinate span width of the antenna positions in meters.
    N_lim : float
        The north coordinate span width of the antenna positions in meters.
    U_lim : float
        The up coordinate span width of the antenna positions in meters.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n_antenna_E x n_antenna_N' antenna locations (x_i, y_i) uniformly distributed, with Up component = U_lim.
    e = np.linspace(-E_lim / 2, E_lim / 2, n_antenna_E)
    n = np.linspace(-N_lim / 2, N_lim / 2, n_antenna_N)
    E, N = np.meshgrid(e, n)
    E_flat = E.flatten()
    N_flat = N.flatten()
    U_flat = np.zeros_like(E_flat) + U_lim
    array_grid = np.column_stack((E_flat, N_flat, U_flat))
    return array_grid


########################################
#  Compute baselines and uv sampling   #
########################################


def get_baselines(array):
    """Get baselines (JAX version).

    Function to compute the baselines of an antenna array.

    Parameters
    ----------
    array : np.ndarray
        The antenna array positions.

    Returns
    -------
    baselines : np.ndarray
        The baselines of the antenna array.
    """
    # Get the baseline for every combination of antennas except for i=j baselines.
    array = jnp.asarray(array)
    diffs = array[:, None, :] - array[None, :, :]  # Shape: (n, n, 3)
    mask = ~jnp.eye(array.shape[0], dtype=bool)  # Shape: (n, n), True where i ≠ j
    return diffs[mask].reshape(-1, 3)


@jit
def ENU_to_XYZ(b_ENU, lat=35.0 / 180 * jnp.pi):
    """ENU to XYZ (JAX version).

    Function to convert the baselines from East-North-Up (ENU) to XYZ coordinates.

    Parameters
    ----------
    b_ENU : np.ndarray
        The baselines in ENU coordinates.
    lat : float
        The latitude of the antenna array in radians.

    Returns
    -------
    X : jnp.ndarray
        The X coordinate of the baselines in XYZ coordinates.
    Y : jnp.ndarray
        The Y coordinate of the baselines in XYZ coordinates.
    Z : jnp.ndarray
        The Z coordinate of the baselines in XYZ coordinates.
    """
    # Compute baseline length, Azimuth and Elevation angles
    D = jnp.sqrt(jnp.sum(b_ENU**2, axis=1))
    A = jnp.arctan2(b_ENU[:, 0], b_ENU[:, 1])
    E = jnp.arcsin(b_ENU[:, 2] / D)
    # Compute the baseline in XYZ coordinates
    X = D * (jnp.cos(lat) * jnp.sin(E) - jnp.sin(lat) * jnp.cos(E) * jnp.cos(A))
    Y = D * jnp.cos(E) * jnp.sin(A)
    Z = D * (jnp.sin(lat) * jnp.sin(E) + jnp.cos(lat) * jnp.cos(E) * jnp.cos(A))

    return X, Y, Z


@jit
def XYZ_to_uvw(X, Y, Z, dec=30.0 / 180 * jnp.pi, ha=0.0, f=1420e6):
    """XYZ to uvw (JAX version).

    Get the uvw sampling points from the XYZ coordinates given a
    source declination, hour angle and frequency.

    Parameters
    ----------
    X : np.ndarray
        The X coordinate of the baselines in XYZ coordinates.
    Y : np.ndarray
        The Y coordinate of the baselines in XYZ coordinates.
    Z : np.ndarray
        The Z coordinate of the baselines in XYZ coordinates.
    dec : float
        The declination of the source in radians.
    ha : float
        The hour angle of the source in radians.
    f : float
        The frequency of the observation in Hz.

    Returns
    -------
    u : np.ndarray
        The u coordinate of the baselines in uvw coordinates.
    v : np.ndarray
        The v coordinate of the baselines in uvw coordinates.
    w : np.ndarray
        The w coordinate of the baselines in uvw coordinates.
    """
    c = 299792458.0
    lam_inv = f / c
    u = lam_inv * (jnp.sin(ha) * X + jnp.cos(ha) * Y)
    v = lam_inv * (
        -jnp.sin(dec) * jnp.cos(ha) * X
        + jnp.sin(dec) * jnp.sin(ha) * Y
        + jnp.cos(dec) * Z
    )
    w = lam_inv * (
        jnp.cos(dec) * jnp.cos(ha) * X
        - jnp.cos(dec) * jnp.sin(ha) * Y
        + jnp.sin(dec) * Z
    )
    return u, v, w


def uv_track_multiband(
    b_ENU,
    lat=35.0 / 180 * jnp.pi,
    dec=35.0 / 180 * jnp.pi,
    track_time=0.0,
    t_0=0.0,
    n_times=1,
    f=1420e6,
    df=0.0,
    n_freqs=1,
    multi_band=False,
):
    """Uv track multiband (JAX version).

    Function to compute the uv sampling baselines for a given observation time and frequency range.

    Parameters
    ----------
    b_ENU : np.ndarray
        The baselines in ENU coordinates.
    lat : float
        The latitude of the antenna array in radians.
    dec : float
        The declination of the source in radians.
    track_time : float
        The duration of the tracking in hours.
    t_0 : float
        The initial tracking time in hours.
    n_times : int
        The number of time steps.
    f : float
        The central frequency of the observation in Hz.
    df : float
        The frequency range of the observation in Hz.
    n_freqs : int
        The number of frequency samples.
    multi_band : bool
        If True separate the uv samples per frequency bands.

    Returns
    -------
    track : jnp.ndarray
        The uv sampling baselines listed for each time step and frequency.
    f_range : jnp.ndarray
        The list of frequency bands used in the simulation.
    """
    # Compute the baselines in XYZ coordinates
    X, Y, Z = ENU_to_XYZ(b_ENU, lat)
    # Compute the time steps
    h = jnp.linspace(t_0, t_0 + track_time, n_times) * jnp.pi / 12
    # Compute the frequency range
    f_range = jnp.linspace(f - df / 2, f + df / 2, n_freqs)

    # Vectorise the function XYZ_to_uvw over the time steps h (5th argument of XYZ_to_uvw)
    vmap_ha = vmap(XYZ_to_uvw, in_axes=(None, None, None, None, 0, None))

    # Vectorise the function vmap_ha over the frequency bands f_range (6th argument of vmap_ha)
    vmap_f = vmap(vmap_ha, in_axes=(None, None, None, None, None, 0))

    # Compute the uvw coordinates for all baselines, time steps and frequency bands
    u, v, w = vmap_f(X, Y, Z, dec, h, f_range)

    # Stack the uvw coordinates in a single array
    track_f = jnp.stack([u, v, w], axis=-1)

    if multi_band:
        # Separate the uv samples per frequency bands
        track = track_f.reshape(n_freqs, -1, 3)
    else:
        # Concatenate all uv samples in a single array
        track = track_f.reshape(-1, 3)
    return track, f_range


def combine_antenna_arr(arr1, arr2):
    """Combine antenna arr.

    Function to combine two antenna arrays.

    Parameters
    ----------
    arr1 : np.ndarray
        The first antenna array positions.
    arr2 : np.ndarray
        The second antenna array positions.

    Returns
    -------
    arr : np.ndarray
        The combined antenna array positions.
    """
    return np.concatenate((arr1, arr2), axis=0)


def load_antenna_enu_txt(path, noise=False):
    """Load antenna txt.

    Function to load antenna name, ENU positions and noise from a txt file.

    Parameters
    ----------
    path : str
        The path to the txt file.
    noise : bool
        Specify if the file contain the noise level of each antenna.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions.
    noise_level : np.ndarray
        If noise==True, return antenna noise level.
    """
    antenna_info = np.genfromtxt(path)
    name = antenna_info[:, 0]

    if noise == True:
        noise_level = antenna_info[:, -1]
        return antenna_info[:, 1:4], noise_level
    else:
        return antenna_info[:, 1:4]


def load_antenna_latlon_txt(path, noise=False):
    """(lat,long,altitude) to ENU.

    Function to load the antenna name, latitude (degree), longitude (degree), altitude (metre) and noise from a txt file,
    and return the information with ENU positions.
    The reference is taken as the middle of all the antennas.
    Waring: the computation of the upp coordinate is valid for antennas that are close enough.

    Parameters
    ----------
    path : str
        The path to the txt file.
    noise : bool
        Specify if the file contain the noise level of each antenna.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions.
    noise_level : np.ndarray
        If noise==True, return antenna noise level.
    """
    R = 6378137.0  # Mean radius of the Earth (m)

    antenna_info = np.genfromtxt(path)
    name = antenna_info[:, 0]
    antenna_pos = antenna_info[:, 1:4]

    # Compute the reference
    ref = np.mean(antenna_pos, axis=0)
    lat0, lon0, h0 = np.radians(ref[0]), np.radians(ref[1]), ref[2]

    enu_list = []
    for lat, lon, h in antenna_pos:
        # Compute the difference between the antenna and the reference (deg, deg, m)
        d_lat = np.radians(lat - ref[0])
        d_lon = np.radians(lon - ref[1])
        d_h = h - h0

        # Convert into ENU
        d_north = d_lat * R
        d_east = d_lon * R * np.cos(lat0)
        d_up = d_h
        enu_list.append((d_east, d_north, d_up))
    enu_array = np.array(enu_list)

    if noise == True:
        noise_level = antenna_info[:, -1]
        return enu_array, noise_level
    else:
        return enu_array


def save_antenna_enu_txt(antenna, path, noise=None):  # pragma: no cover
    """Save the antenna information into a txt file.

    Function to save the antenna name, ENU positions and optionally noise into a txt file.

    Parameters
    ----------
    antenna : np.ndarray
        The numpy array which contain the antenna information.
    path : str
        The path to the txt file.
    noise  np.ndarray
        The numpy array which contain the noise information, if available.

    Returns
    -------
    None
        The function only saves the file.
    """
    nb_antennas = len(antenna)
    antenna_name = np.linspace(1, nb_antennas, nb_antennas, dtype=int).reshape(-1, 1)

    if noise is not None:
        antenna_info = np.concatenate(
            (antenna_name, antenna, noise.reshape(-1, 1)), axis=1
        )
        fmt = "%d %.6f %.6f %.6f %.6f"
        header = "Name | E | N | U | Noise level"
    else:
        antenna_info = np.concatenate((antenna_name, antenna), axis=1)
        fmt = "%d %.6f %.6f %.6f"
        header = "Name | E | N | U"

    np.savetxt(
        path,
        antenna_info,
        fmt=fmt,
        header=header,
    )



import numpy as np
import numpy.testing as npt

from argosim import antenna_utils as au


class TestAntennaUtils:

    antenna_lims = {"east": (-500.0, 500.0), "north": (-500.0, 500.0), "up": 0.0}

    circ_antenna_exp = np.array(
        [[300.0, 0.0, 0.0], [-150.0, 259.80762114, 0.0], [-150.0, -259.80762114, 0.0]]
    )
    circ_antenna_tolerance = 1e-7

    y_antenna_exp = np.array(
        [
            [100.0, 0.0, 0.0],
            [200.0, 0.0, 0.0],
            [300.0, 0.0, 0.0],
            [400.0, 0.0, 0.0],
            [500.0, 0.0, 0.0],
            [-50.0, 86.60254038, 0.0],
            [-100.0, 173.20508076, 0.0],
            [-150.0, 259.80762114, 0.0],
            [-200.0, 346.41016151, 0.0],
            [-250.0, 433.01270189, 0.0],
            [-50.0, -86.60254038, 0.0],
            [-100.0, -173.20508076, 0.0],
            [-150.0, -259.80762114, 0.0],
            [-200.0, -346.41016151, 0.0],
            [-250.0, -433.01270189, 0.0],
        ]
    )
    y_antenna_tolerance = 1e-7

    random_antenna_seed = 123
    random_antenna_exp = np.array(
        [
            [196.4691856, -213.86066505, 0.0],
            [51.31476908, 219.46896979, 0.0],
            [480.76419838, 184.82973858, 0.0],
        ]
    )
    random_antenna_tolerance = 1e-7

    uni_antenna_span = {"east": 800.0, "north": 800.0, "up": 0.0}
    uni_antenna_test_element = 333
    uni_antenna_333 = np.array([-64.51612903, -141.93548387, 0.0])
    uni_antenna_tolerance = 1e-7

    ENU_antenna_in = np.array([[101.0, 200.5, -33.0]])
    XYZ_antenna_exp = {"X": -142.03409295, "Y": 101.0, "Z": 145.31196248}
    ENU_to_XYZ_decimal_tol = 4
    XYZ_to_uvw_decimal_tol = 4
    uvw_exp = {"u": 478.39762533, "v": 932.45332539, "w": -238.48430193}

    enu_txt_arr_path = "configs/arrays/argos_pathfinder.enu.txt"
    enu_txt_arr_exp = np.array(
        [
            [15.0, 10.0, 0.0],
            [17.0, 35.0, 0.0],
            [7.0, 50.0, 0.0],
            [17.0, 75.0, 0.0],
            [10.0, 183.0, 0.0],
        ]
    )
    latlon_txt_arr_path = "configs/arrays/array_lat_lon.txt"
    latlon_txt_arr_exp = np.array(
        [[-18159.17012419, -5565.97453966, 0.0], [18159.17012419, 5565.97453966, 0.0]]
    )

    random_antenna_baselines_exp = np.array(
        [
            [145.15441652, -433.32963484, 0.0],
            [-284.29501278, -398.69040363, 0.0],
            [-145.15441652, 433.32963484, 0.0],
            [-429.4494293, 34.63923121, 0.0],
            [284.29501278, 398.69040363, 0.0],
            [429.4494293, -34.63923121, 0.0],
        ]
    )

    uv_track_default_exp = np.array(
        [
            [6.87539883e02, -2.05251355e03, -1.34622557e-13],
            [-1.34659464e03, -1.88844101e03, -1.34622557e-13],
            [-6.87539883e02, 2.05251355e03, 1.34622557e-13],
            [-2.03413453e03, 1.64072534e02, 0.00000000e00],
            [1.34659464e03, 1.88844101e03, 0.00000000e00],
            [2.03413453e03, -1.64072534e02, 1.68278197e-14],
        ]
    )
    uv_track_params = (
        35.0 / 180 * np.pi,  # lat
        35.0 / 180 * np.pi,  # dec
        1.5,  # track_time
        -0.5,  # t_0
        3,  # n_times
        1420e6,  # f
        100.0e6,  # df
        2,  # n_freqs
        False,  # multi_band
    )
    uv_track_exp = np.array(
        [
            [509.40157076, -2024.3298598, 62.96406953],
            [-1424.4678981, -1719.55331102, -146.23278783],
            [-509.40157076, 2024.3298598, -62.96406953],
            [-1933.86946886, 304.77654878, -209.19685736],
            [1424.4678981, 1719.55331102, 146.23278783],
            [1933.86946886, -304.77654878, 209.19685736],
            [736.19667995, -1953.96308977, -37.53009284],
            [-1228.04974932, -1869.40026143, 67.77083571],
            [-736.19667995, 1953.96308977, 37.53009284],
            [-1964.24642927, 84.56282834, 105.30092855],
            [1228.04974932, 1869.40026143, -67.77083571],
            [1964.24642927, -84.56282834, -105.30092855],
            [934.70016361, -1859.57020356, -172.33710512],
            [-984.43833736, -1994.38890869, 246.27312317],
            [-934.70016361, 1859.57020356, 172.33710512],
            [-1919.13850097, -134.81870513, 418.6102283],
            [984.43833736, 1994.38890869, -246.27312317],
            [1919.13850097, 134.81870513, -418.6102283],
            [546.58416716, -2172.09116343, 67.55998701],
            [-1528.44365709, -1845.06815124, -156.90671395],
            [-546.58416716, 2172.09116343, -67.55998701],
            [-2075.02782425, 327.0230122, -224.46670096],
            [1528.44365709, 1845.06815124, 156.90671395],
            [2075.02782425, -327.0230122, 224.46670096],
            [789.93366389, -2096.58813282, -40.26951567],
            [-1317.68841715, -2005.85283525, 72.71761204],
            [-789.93366389, 2096.58813282, 40.26951567],
            [-2107.62208104, 90.73529756, 112.98712771],
            [1317.68841715, 2005.85283525, -72.71761204],
            [2107.62208104, -90.73529756, -112.98712771],
            [1002.92645292, -1995.30525491, -184.91645586],
            [-1056.29515031, -2139.96474145, 264.24926355],
            [-1002.92645292, 1995.30525491, 184.91645586],
            [-2059.22160323, -144.65948653, 449.16571942],
            [1056.29515031, 2139.96474145, -264.24926355],
            [2059.22160323, 144.65948653, -449.16571942],
        ]
    )
    multiband_track_shape_exp = (2, 18, 3)

    uv_atol = 1e-3

    def test_random_antenna_pos_default(self):

        antenna_pos = au.random_antenna_pos()

        assert np.logical_and(
            antenna_pos[0] >= self.antenna_lims["east"][0],
            antenna_pos[0] <= self.antenna_lims["east"][1],
        )
        assert np.logical_and(
            antenna_pos[1] >= self.antenna_lims["north"][0],
            antenna_pos[1] <= self.antenna_lims["north"][1],
        )
        assert antenna_pos[2] == self.antenna_lims["up"]

    def test_circular_antenna_arr(self):

        circ_antenna_out = au.circular_antenna_arr()

        npt.assert_allclose(
            circ_antenna_out,
            self.circ_antenna_exp,
            atol=self.circ_antenna_tolerance,
            err_msg="Circular antenna outputs do not match.",
        )

    def test_y_antenna_arr(self):

        y_antenna_out = au.y_antenna_arr()

        npt.assert_allclose(
            y_antenna_out,
            self.y_antenna_exp,
            atol=self.y_antenna_tolerance,
            err_msg="Y antenna array outputs do not match.",
        )

    def test_random_antenna_arr(self):

        random_antenna_out = au.random_antenna_arr(seed=self.random_antenna_seed)

        npt.assert_allclose(
            random_antenna_out,
            self.random_antenna_exp,
            atol=self.random_antenna_tolerance,
            err_msg="Random antenna array outputs do not match.",
        )

    def test_uni_antenna_arr(self):

        uni_antenna_out = au.uni_antenna_array()
        uni_antenna_span_out = uni_antenna_out[-1] - uni_antenna_out[0]

        assert uni_antenna_span_out[0] == self.uni_antenna_span["east"]
        assert uni_antenna_span_out[1] == self.uni_antenna_span["north"]
        assert uni_antenna_span_out[2] == self.uni_antenna_span["up"]

        npt.assert_allclose(
            self.uni_antenna_333,
            uni_antenna_out[self.uni_antenna_test_element],
            atol=self.uni_antenna_tolerance,
            err_msg="Uniform antenna output does not match at position {}.".format(
                self.uni_antenna_test_element
            ),
        )

    def test_enu_to_xyz(self):
        x, y, z = au.ENU_to_XYZ(self.ENU_antenna_in)
        npt.assert_almost_equal(
            x,
            self.XYZ_antenna_exp["X"],
            decimal=self.ENU_to_XYZ_decimal_tol,
            err_msg="X coordinate conversion from ENU to XYZ does not match.",
        )
        npt.assert_almost_equal(
            y,
            self.XYZ_antenna_exp["Y"],
            decimal=self.ENU_to_XYZ_decimal_tol,
            err_msg="Y coordinate conversion from ENU to XYZ does not match.",
        )
        npt.assert_almost_equal(
            z,
            self.XYZ_antenna_exp["Z"],
            decimal=self.ENU_to_XYZ_decimal_tol,
            err_msg="Z coordinate conversion from ENU to XYZ does not match.",
        )

    def test_xyz_to_uvw(self):
        dict = self.XYZ_antenna_exp
        x, y, z = dict["X"], dict["Y"], dict["Z"]
        u, v, w = au.XYZ_to_uvw(x, y, z)

        npt.assert_almost_equal(
            u,
            self.uvw_exp["u"],
            decimal=self.XYZ_to_uvw_decimal_tol,
            err_msg="U coordinate conversion from XYZ to UVW does not match.",
        )
        npt.assert_almost_equal(
            v,
            self.uvw_exp["v"],
            decimal=self.XYZ_to_uvw_decimal_tol,
            err_msg="V coordinate conversion from XYZ to UVW does not match.",
        )
        npt.assert_almost_equal(
            w,
            self.uvw_exp["w"],
            decimal=self.XYZ_to_uvw_decimal_tol,
            err_msg="W coordinate conversion from XYZ to UVW does not match.",
        )

    def test_combine_antenna_arr(self):
        arr1 = au.random_antenna_arr()
        arr2 = au.random_antenna_arr()
        arr = au.combine_antenna_arr(arr1, arr2)
        n_antenna_exp = au.random_antenna_arr.__defaults__[0] * 2

        assert (
            len(arr) == n_antenna_exp
        ), "Combined antenna array length does not match expected."

    def test_load_antenna_enu_txt(self):
        arr = au.load_antenna_enu_txt(self.enu_txt_arr_path)
        npt.assert_allclose(
            arr,
            self.enu_txt_arr_exp,
            err_msg="Antenna array loaded from ENU text file does not match expected output.",
        )

    def test_load_antenna_latlon_txt(self):
        arr = au.load_antenna_latlon_txt(self.latlon_txt_arr_path)
        npt.assert_allclose(
            arr,
            self.latlon_txt_arr_exp,
            err_msg="Antenna array loaded from latitude/longitude text file does not match expected output.",
        )

    def test_get_baselines(self):
        baselines_out = au.get_baselines(self.random_antenna_exp)
        npt.assert_allclose(
            baselines_out,
            self.random_antenna_baselines_exp,
            atol=self.uv_atol,
            err_msg="Baselines computed from random antenna array do not match expected output.",
        )

    def test_uv_track_default(self):
        track, _ = au.uv_track_multiband(self.random_antenna_baselines_exp)
        npt.assert_allclose(
            track,
            self.uv_track_default_exp,
            atol=self.uv_atol,
            err_msg="UV track computed from random antenna baselines does not match expected output.",
        )

    def test_uv_track_params(self):
        track, _ = au.uv_track_multiband(
            self.random_antenna_baselines_exp, *self.uv_track_params
        )
        npt.assert_allclose(
            track,
            self.uv_track_exp,
            atol=self.uv_atol,
            err_msg="UV track computed from random antenna baselines does not match expected output.",
        )

    def test_uv_track_multiband(self):
        track, _ = au.uv_track_multiband(
            self.random_antenna_baselines_exp,
            *self.uv_track_params[:-1],
            multi_band=True,
        )
        track_shape_out = track.shape
        npt.assert_equal(
            track_shape_out,
            self.multiband_track_shape_exp,
            err_msg="Multiband UV track shape does not match expected output.",
        )



import numpy as np
import numpy.testing as npt

import argosim.beam_utils as abu


class TestBeamUtils:
    beam_uv_coords = (3.3, -10.5)
    beam_value_exp = -0.6757983092461403
    beam_value_exp_decimal = 10

    r_fov_exp = 1.9644846081521368
    r_fov_exp_decimal = 10

    beam_edge_exp = 6.544984694978735
    beam_edge_exp_decimal = 10

    fov_solid_angle_exp = 12.124033663969739
    fov_solid_angle_exp_decimal = 10

    beam_cos_cube_path = "src/argosim/tests/data/beam_cos_cube.npy"

    def test_call(self):
        beam = abu.CosCubeBeam()
        val_out = beam(*self.beam_uv_coords)
        npt.assert_almost_equal(
            val_out,
            self.beam_value_exp,
            decimal=self.beam_value_exp_decimal,
            err_msg="Beam value does not match expected value.",
        )

    def test_r_fov(self):
        beam = abu.CosCubeBeam()
        r_fov_out = beam.r_fov()
        npt.assert_almost_equal(
            r_fov_out,
            self.r_fov_exp,
            decimal=self.r_fov_exp_decimal,
            err_msg="FOV radius does not match expected value.",
        )

    def test_beam_edge(self):
        beam = abu.CosCubeBeam()
        beam_edge_out = beam.beam_edge()
        npt.assert_almost_equal(
            beam_edge_out,
            self.beam_edge_exp,
            decimal=self.beam_edge_exp_decimal,
            err_msg="Beam edge does not match expected value.",
        )

    def test_fov_solid_angle(self):
        beam = abu.CosCubeBeam()
        fov_solid_angle_out = beam.fov_solid_angle()
        npt.assert_almost_equal(
            fov_solid_angle_out,
            self.fov_solid_angle_exp,
            decimal=self.fov_solid_angle_exp_decimal,
            err_msg="FOV solid angle does not match expected value.",
        )

    def test_get_beam(self):
        beam = abu.CosCubeBeam()
        beam_image_out = beam.get_beam()
        beam_image_exp = np.load(self.beam_cos_cube_path)
        npt.assert_almost_equal(
            beam_image_out,
            beam_image_exp,
            decimal=self.beam_value_exp_decimal,
            err_msg="Beam image does not match expected value.",
        )



"""Plot utils.

This module contains functions to plot the antenna array, beam, baselines,
uv-coverage and sky models.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Ellipse
from skimage.transform import rotate

from argosim import metrics_utils


def plot_antenna_arr(
    array, ax=None, fig=None, title="Array", antenna_idx=True, s=20, c="mediumblue"
):  # pragma: no cover
    """Plot antenna array.

    Function to plot the antenna array in ground coordinates.

    Parameters
    ----------
    array : np.ndarray
        The antenna array positions in the ground.
    ax : matplotlib.axes.Axes
        The axis to plot the antenna array. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the antenna array. For plotting on a specific subplot axis.
    title : str
        The title of the plot.
    antenna_idx : bool
        If True, annotate the antenna indices on the plot.
    s : int
        The size of the scatter points. Default is 20.
    c : str
        The color of the scatter points. Default is "mediumblue".

    Returns
    -------
    None
    """
    # Center the array
    array = array - np.mean(array, axis=0)
    # Convert to km
    array = array / 1000.0
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    ax.scatter(array[:, 0], array[:, 1], s=s, c=c)
    if antenna_idx:
        for i, txt in enumerate(range(1, len(array) + 1, 1)):
            ax.annotate(txt, (array[i, 0], array[i, 1]))
    ax.set_xlabel("E [km]")
    ax.set_ylabel("N [km]")
    x_lim = np.max(np.abs(array)) * 1.1
    y_lim = x_lim
    ax.set_xlim(-x_lim, x_lim)
    ax.set_ylim(-y_lim, y_lim)
    ax.set_aspect("equal", adjustable="box")
    ax.set_title(title)
    if ax is None or fig is None:
        plt.show()


def plot_baselines(
    baselines, ax=None, fig=None, title=None, ENU=False, s=5, c="darkred"
):  # pragma: no cover
    """Plot baselines.

    Function to plot the baselines in uv-space.

    Parameters
    ----------
    baselines : np.ndarray
        The uv-space sampling positions.
    ax : matplotlib.axes.Axes
        The axis to plot the baselines. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the baselines. For plotting on a specific subplot axis.
    title : str
        The title of the plot. If None, a default title is used.
    ENU : bool
        If True, plot the baselines in East-North-Up coordinates. Otherwise, plot in uv-space.
    s : int
        The size of the scatter points. Default is 5.
    c : str
        The color of the scatter points. Default is "darkred".

    Returns
    -------
    None
    """
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    # Convert to km
    baselines = baselines / 1000.0
    if ENU:
        ax.set_xlabel("East [km]")
        ax.set_ylabel("North [km]")
        if title is None:
            ax.set_title("Baselines")
        else:
            ax.set_title(title)
        ax.scatter(baselines[:, 0], baselines[:, 1], s=s, c=c)
    else:
        ax.set_xlabel(r"$u$(k$\lambda$)")
        ax.set_ylabel(r"$v$(k$\lambda$)")
        if title is None:
            ax.set_title(r"uv-plane")
        else:
            ax.set_title(title)
        ax.scatter(baselines[:, 0], baselines[:, 1], s=s, c=c)
    ax.set_xlim([np.min(baselines), np.max(baselines)])
    ax.set_ylim([np.min(baselines), np.max(baselines)])
    ax.set_aspect("equal", adjustable="box")
    if ax is None or fig is None:
        plt.show()


def plot_sky(
    image, fov_size, ax=None, fig=None, title=None, cbar=True
):  # pragma: no cover
    """Plot sky.

    Function to plot the sky model.

    Parameters
    ----------
    image : np.ndarray
        The image to be plotted in real space.
    fov_size : tuple
        The field of view size in degrees.
    ax : matplotlib.axes.Axes
        The axis to plot the sky model. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the sky model. For plotting on a specific subplot axis.
    title : str
        The title of the plot. Default is "Sky".
    cbar : bool
        If True, display a colorbar. Default is True.

    Returns
    -------
    None
    """
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    im = ax.imshow(
        image,
        extent=[-fov_size[0] / 2, fov_size[0] / 2, -fov_size[1] / 2, fov_size[1] / 2],
        origin="lower",
    )
    if cbar:
        fig.colorbar(im, ax=ax)
    ax.set_xlabel("l [deg]")
    ax.set_ylabel("m [deg]")
    if title is None:
        ax.set_title("{} ({}x{})".format("Sky", image.shape[0], image.shape[1]))
    else:
        ax.set_title(title)
    if ax is None or fig is None:
        plt.show()


def plot_sky_uv(
    image_uv, fov_size, ax=None, fig=None, title="Sky uv", cbar=False, scale="linear"
):  # pragma: no cover
    """Plot sky uv.

    Function to plot the absolute value of an image in uv-space.

    Parameters
    ----------
    image_uv : np.ndarray
        The image to be plotted in uv-space.
    fov_size : tuple
        The field of view size in degrees.
    ax : matplotlib.axes.Axes
        The axis to plot the sky model in uv-space. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the sky model in uv-space. For plotting on a specific subplot axis.
    title : str
        The title of the plot. Default is "Sky uv".
    cbar : bool
        If True, display a colorbar. Default is False.
    scale : str
        The scale of the uv-image. Default is 'linear'. Options are 'linear' or 'log'.

    Returns
    -------
    None
    """
    max_u = (180 / np.pi) * image_uv.shape[0] / (2 * fov_size[0]) / 1000
    max_v = (180 / np.pi) * image_uv.shape[1] / (2 * fov_size[1]) / 1000

    if ax is None or fig is None:
        fig, ax = plt.subplots(1, 1)
    if scale == "log":
        image_uv = np.log10(np.abs(image_uv) + 1e-10)
    elif scale == "linear":
        image_uv = np.abs(image_uv)
    im = ax.imshow(image_uv, extent=[-max_u, max_u, -max_v, max_v], origin="lower")
    if cbar:
        fig.colorbar(im, ax=ax)
    ax.set_xlabel(r"$u$(k$\lambda$)")
    ax.set_ylabel(r"$v$(k$\lambda$)")
    ax.set_title("Amplitude")
    if ax is None or fig is None:
        plt.show()


def plot_uv_hist(baselines, bins=20, output_folder=None):  # pragma: no cover
    """Plot uv histogram.

    Function to plot the histogram of the uv-sampling distribution.

    Parameters
    ----------
    baselines : np.ndarray
        The uv-space sampling positions.
    bins : int
        The number of bins for the histogram.
    output_folder : str
        The output folder to save the plot.

    Returns
    -------
    np.ndarray
        The histogram of the uv-sampling distribution.
    """
    # scale to kilo-lambda
    baselines = baselines / 1000

    cmap = matplotlib.colormaps["bone"]

    fig, ax = plt.subplots(1, 2, figsize=(11, 4))

    D = np.sqrt(np.sum(baselines[:, :2] ** 2, axis=1))
    baseline_hist = ax[0].hist(D, range=(0, np.max(D) * 1.1), bins=bins)

    n = baseline_hist[0]
    patches = baseline_hist[2]
    col = (n - n.min()) / (n.max() - n.min())
    for c, p in zip(col, patches):
        plt.setp(p, "facecolor", cmap(c))
    ax[0].set_title("Baselines histogram")
    ax[0].set_xlabel(r"UV distance $(k\lambda)$")
    ax[0].set_ylabel("Counts")
    ax[0].set_facecolor("palegoldenrod")
    ax[0].set_box_aspect(1)

    counts = np.flip(baseline_hist[0])
    r_list = baseline_hist[1]

    colors = cmap((counts / max(counts)))

    draw_back = plt.Circle((0.0, 0.0), 10 * r_list[-1], color="black", fill=True)
    ax[1].add_artist(draw_back)
    for color, r in zip(colors, np.flip(r_list[1:])):
        draw1 = plt.Circle((0.0, 0.0), r, color=color, fill=True)
        ax[1].add_artist(draw1)

    ax[1].scatter(baselines[:, 0], baselines[:, 1], color="yellow", s=1, alpha=0.3)

    fig.colorbar(
        plt.cm.ScalarMappable(
            cmap=cmap, norm=matplotlib.colors.Normalize(vmin=0, vmax=max(counts))
        ),
        ax=ax[1],
        orientation="vertical",
        label="Counts",
    )

    ax[1].set_aspect("equal")
    ax[1].set_xlim(-r_list[-1] * 1.1, r_list[-1] * 1.1)
    ax[1].set_ylim(-r_list[-1] * 1.1, r_list[-1] * 1.1)
    ax[1].set_title("Radial distribution")
    ax[1].set_xlabel(r"u $(k\lambda)$")
    ax[1].set_ylabel(r"v $(k\lambda)$")

    plt.suptitle(
        "UV sampling distribution",
        horizontalalignment="center",
        verticalalignment="top",
    )

    if output_folder is not None:
        plt.savefig(output_folder + "uv_hist.pdf")
    else:
        plt.show()

    return baseline_hist


def plot_beam_fit(beam, fit_result=None):  # pragma: no cover
    """Plot beam fit.

    Plot the beam and overlay the fitted ellipse.

    Parameters
    ----------
    beam : np.ndarray
        2D beam image.
    fit_result : dict
        Dictionary containing the ellipse parameters (center, width, height, angle_deg, eccentricity). If None, it is computed from the beam.

    Returns
    -------
    None
    """
    if fit_result is None:
        fit_result = metrics_utils.fit_elliptical_beam(beam)

    plt.imshow(beam, origin="lower")
    plt.colorbar()

    ellipse = Ellipse(
        xy=fit_result["center"],
        width=fit_result["width"],
        height=fit_result["height"],
        angle=fit_result["angle_deg"],
        edgecolor="red",
        facecolor="none",
        lw=2,
        label="Fitted Ellipse",
    )

    ax = plt.gca()
    ax.add_patch(ellipse)
    ax.set_aspect("equal")
    plt.title(f"Elliptical Fit (e = {fit_result['eccentricity']:.4f})")
    plt.legend()
    plt.show()


def plot_beam_and_fit(beam, fov_size, fit_result=None):  # pragma: no cover
    """Plot beam and fit.

    Plot the beam and overlay the fitted ellipse.

    Parameters
    ----------
    beam : np.ndarray
        2D beam image.
    fov_size : tuple
        The field of view size in degrees.
    fit_result : dict
        Dictionary containing the ellipse parameters (center, width, height, angle_deg, eccentricity). If None, it is computed from the beam.

    Returns
    -------
    None
    """
    if fit_result is None:
        fit_result = metrics_utils.fit_elliptical_beam(beam)
    Npx = beam.shape[0]
    # Beam fit ellipse
    ellipse = Ellipse(
        xy=fit_result["center"],
        width=fit_result["width"],
        height=fit_result["height"],
        angle=fit_result["angle_deg"],
        edgecolor="k",
        facecolor="none",
        lw=2,
        label="Elliptical fit",
    )
    # Semi-major and semi-minor beam gain
    beam_rotated = rotate(beam, fit_result["angle_deg"], resize=True)
    semi_major_beam = beam_rotated[
        beam_rotated.shape[0] // 2,
        beam_rotated.shape[0] // 2 - Npx // 2 : beam_rotated.shape[0] // 2 + Npx // 2,
    ]
    semi_minor_beam = beam_rotated[
        beam_rotated.shape[0] // 2 - Npx // 2 : beam_rotated.shape[0] // 2 + Npx // 2,
        beam_rotated.shape[1] // 2,
    ]
    # First plot: beam and fit, second plot: semi-major beam gain, third plot: semi-minor beam gain.
    fig, ax = plt.subplots(1, 3, figsize=(25, 10))
    ax[0].imshow(beam, origin="lower")
    ax[0].set_yticks(
        [Npx // 10, Npx // 2, Npx - Npx // 10],
        [-fov_size[0] * 2 / 5, 0, fov_size[0] * 2 / 5],
    )
    ax[0].set_xticks(
        [Npx // 10, Npx // 2, Npx - Npx // 10],
        [-fov_size[0] * 2 / 5, 0, fov_size[0] * 2 / 5],
    )
    ax[0].set_xlabel("l (deg)")
    ax[0].set_ylabel("m (deg)")
    # Plot the ellipse
    ax[0].add_patch(ellipse)
    # Plot the semi-major and semi-minor axes as lines through the center of the ellipse
    ax[0].plot(
        [0, Npx - 1],
        [
            Npx // 2 - Npx // 2 * np.tan(np.radians(fit_result["angle_deg"])),
            Npx // 2 + Npx // 2 * np.tan(np.radians(fit_result["angle_deg"])),
        ],
        color="orange",
        alpha=0.7,
        lw=3,
        label="Major axis",
        linestyle="--",
    )
    ax[0].plot(
        [
            Npx // 2 - Npx // 2 * np.tan(np.radians(fit_result["angle_deg"])),
            Npx // 2 + Npx // 2 * np.tan(np.radians(fit_result["angle_deg"])),
        ],
        [Npx - 1, 0],
        color="royalblue",
        alpha=0.7,
        lw=4,
        label="Minor axis",
        linestyle="--",
    )
    ax[0].set_title("Dirty beam")
    ax[0].legend(loc="upper right", fontsize=32)

    # Plot the semi-major and semi-minor beam gain
    ax[1].plot(semi_major_beam, color="orange", lw=4)
    ax[1].set_title("Beam gain major axis")
    ax[1].set_yticks([])
    ax[1].set_box_aspect(1)
    ax[1].set_xticks(
        [Npx // 10, Npx // 2, Npx - Npx // 10],
        [-fov_size[0] * 2 / 5, 0, fov_size[0] * 2 / 5],
    )
    ax[1].set_xlabel("(deg)")
    ax[1].plot(
        [int(Npx // 2 - fit_result["width"]), int(Npx // 2 + fit_result["width"])],
        [np.max(semi_major_beam) / 2, np.max(semi_major_beam) / 2],
        color="black",
        linestyle="--",
        lw=4,
        label="FWHM",
    )
    ax[1].scatter(
        [
            int(Npx // 2 - fit_result["width"] + 1),
            int(Npx // 2 + fit_result["width"] + 1),
        ],
        [np.max(semi_major_beam) / 2, np.max(semi_major_beam) / 2],
        color="black",
        marker="x",
        s=200,
    )
    ax[1].legend(loc="upper right", fontsize=30)

    ax[2].plot(semi_minor_beam, color="royalblue", lw=4)
    ax[2].set_title("Beam gain minor axis")
    ax[2].set_yticks([])
    ax[2].set_box_aspect(1)
    ax[2].set_xticks(
        [Npx // 10, Npx // 2, Npx - Npx // 10],
        [-fov_size[0] * 2 / 5, 0, fov_size[0] * 2 / 5],
    )
    ax[2].set_xlabel("(deg)")
    ax[2].plot(
        [int(Npx // 2 - fit_result["height"]), int(Npx // 2 + fit_result["height"])],
        [np.max(semi_minor_beam) / 2, np.max(semi_minor_beam) / 2],
        color="black",
        linestyle="--",
        lw=4,
        label="FWHM",
    )
    ax[2].scatter(
        [
            int(Npx // 2 - fit_result["height"] + 1),
            int(Npx // 2 + fit_result["height"] + 1),
        ],
        [np.max(semi_minor_beam) / 2, np.max(semi_minor_beam) / 2],
        color="black",
        marker="x",
        s=200,
    )
    ax[2].legend(loc="upper right", fontsize=30)

    plt.tight_layout()
    plt.show()



import numpy as np
import numpy.testing as npt

import argosim.data_utils as adu


class TestDataUtils:

    gauss_source_path = "src/argosim/tests/data/test_gauss_source.npy"
    gauss_source_decimal = 6

    seed = 24
    cov_expected = np.array([[5.9600173, 2.91338124], [2.91338124, 5.69951205]])
    cov_decimal = 6

    mu_expected = np.array([0.92003461, 0.3990241])
    mu_decimal = 6

    random_source_path = "src/argosim/tests/data/test_random_source.npy"
    random_source_decimal = 5

    sky_model_path = "src/argosim/tests/data/sky_model_exp.npy"
    sky_model_decimal = 5

    def test_gauss_source(self):
        gauss_source_expected = np.load(self.gauss_source_path)
        gauss_source_out = adu.gauss_source()
        npt.assert_almost_equal(
            gauss_source_out,
            gauss_source_expected,
            decimal=self.gauss_source_decimal,
            err_msg="Gaussian source image does not match expected value.",
        )

    def test_sigma2d(self):
        cov_out = adu.sigma2d(seed=self.seed)
        npt.assert_almost_equal(
            cov_out,
            self.cov_expected,
            decimal=self.cov_decimal,
            err_msg="Covariance matrix does not match expected value.",
        )

    def test_mu2d(self):
        mu_out = adu.mu2d(seed=self.seed)
        npt.assert_almost_equal(
            mu_out,
            self.mu_expected,
            decimal=self.mu_decimal,
            err_msg="Mean vector does not match expected value.",
        )

    def test_random_source(self):
        random_source_expected = np.load(self.random_source_path)
        random_source_out = adu.random_source((256, 256), 5, seed=self.seed)
        npt.assert_almost_equal(
            random_source_out,
            random_source_expected,
            decimal=self.random_source_decimal,
            err_msg="Random source image does not match expected value.",
        )

    def test_n_source_sky(self):
        sky_model_expected = np.load(self.sky_model_path)
        sky_model_out = adu.n_source_sky(
            (256, 256), 1.0, [0.01, 0.02, 0.03], [0.4, 0.3, 0.3], seed=332
        )
        npt.assert_almost_equal(
            sky_model_out,
            sky_model_expected,
            decimal=self.sky_model_decimal,
            err_msg="Number of sources in sky model does not match expected value.",
        )



import numpy as np
import numpy.testing as npt

import argosim.clean as ac


class TestClean:

    beam_path = "src/argosim/tests/data/dirty_beam_sim_single_band.npy"
    shifted_beams_path = "src/argosim/tests/data/shifted_beams.npy"

    peaks_exp = [
        (0.00384521484375, 128, 128, 0, 0),
        (0.00384521484375, 228, 238, 100, 110),
        (0.00384521484375, 108, 208, -20, 80),
        (0.00384521484375, 26, 78, -102, -50),
        (0.00384521484375, 160, 108, 32, -20),
    ]

    pad_odd_exp = np.array(
        [
            [1.0, 1.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    obs_path = "src/argosim/tests/data/obs_sim_single_band.npy"
    clean_obs_exp_path = "src/argosim/tests/data/clean_observation.npy"
    clean_obs_res_exp_path = "src/argosim/tests/data/clean_observation_res.npy"
    sky_model_clean_exp_path = "src/argosim/tests/data/sky_model_clean.npy"
    sky_model_clean_res_exp_path = "src/argosim/tests/data/sky_model_clean_res.npy"
    clean_decimal = 7

    def test_shift_beam(self):
        shifted_beams = np.load(self.shifted_beams_path, allow_pickle=True).item()
        shfits = shifted_beams["shifts"]
        beam_exp = shifted_beams["shifted_beams"]
        beam = np.load(self.beam_path)

        for shift, shifted_beam in zip(shfits, beam_exp):
            npt.assert_array_equal(
                ac.shift_beam(beam, shift[0], shift[1]),
                shifted_beam,
                err_msg=f"Shift {shift} did not match expected beam.",
            )

    def test_find_peak(self):
        beams = np.load(self.shifted_beams_path, allow_pickle=True).item()[
            "shifted_beams"
        ]
        peaks_out = [ac.find_peak(beam) for beam in beams]
        npt.assert_array_equal(
            peaks_out,
            self.peaks_exp,
            err_msg="Peaks found did not match expected values.",
        )

    def test_pad_odd(self):
        # Odd array
        pad_in = np.ones((5, 5))
        pad_out = ac.pad_odd(pad_in)
        npt.assert_array_equal(
            pad_out,
            pad_in,
            err_msg="Padding an odd array did not return the same array.",
        )
        # Even array
        pad_in = np.ones((4, 4))
        pad_out = ac.pad_odd(pad_in)
        pad_exp = self.pad_odd_exp
        npt.assert_array_equal(
            pad_out,
            pad_exp,
            err_msg="Padding an even array did not return the expected padded array.",
        )

    def test_clean_hogbom(self):
        obs = np.load(self.obs_path)
        beam = np.load(self.beam_path)
        # Without residuals
        I_clean, sky_model = ac.clean_hogbom(
            obs, beam, 0.3, 100, 1e-2, clean_beam_size_px=10
        )
        npt.assert_array_almost_equal(
            I_clean,
            np.load(self.clean_obs_exp_path),
            decimal=self.clean_decimal,
            err_msg="Cleaned observation did not match expected values.",
        )

        npt.assert_array_almost_equal(
            sky_model,
            np.load(self.sky_model_clean_exp_path),
            decimal=self.clean_decimal,
            err_msg="Sky model from clean did not match expected values.",
        )

        # With residuals
        I_clean_res, sky_model_res = ac.clean_hogbom(
            obs, beam, 0.3, 100, 1e-2, clean_beam_size_px=10, res=True
        )
        npt.assert_array_almost_equal(
            I_clean_res,
            np.load(self.clean_obs_res_exp_path),
            decimal=self.clean_decimal,
            err_msg="Cleaned observation with res==True did not match expected values.",
        )

        npt.assert_array_almost_equal(
            sky_model_res,
            np.load(self.sky_model_clean_res_exp_path),
            decimal=self.clean_decimal,
            err_msg="Sky model with res==True from clean did not match expected values.",
        )



"""Random utils.

A small module to hold a class for a temporary seed,
so as to not mess with any global seed.

:Authors: Samuel Gullin <gullin@ia.forth.gr>

"""

import numpy as np


class local_seed:
    """
    Holds a seed for use with `numpy.random` in a local scope.

    Can be instantiated with no seed, in which case it will yield the default behaviour.
    Note that if `seed` is set to the same seed `numpy` was using outside this scope,
    that doesn't mean the state is the same -- if you intend to keep the same state,
    use `seed = None`.

    Attributes
    ----------
    seed: int, optional
        The seed used for numpy.random in the local scope.
    old_state:
        The state of numpy.random outside this scope, to be resumed afterwards.
    """

    def __init__(self, seed=None):
        """Instantiate with a set seed."""
        self.seed = seed
        self.old_state = None

    def __enter__(self):
        """
        Enter local scope.

        Grabs `numpy.randoms`'s state only if a `seed != None` was supplied
        upon creation, and then seeds `numpy.random`.
        """
        if self.seed:
            self.old_state = np.random.get_state()
            np.random.seed(self.seed)

    def __exit__(self, *_):
        """
        Exit local scope.

        Reverts `numpy.random` to the outer scope's state, if a `seed != None`
        was supplied upon creation.
        """
        if self.seed:
            np.random.set_state(self.old_state)



import numpy.random as rnd

from argosim.rand_utils import local_seed as ls


def test_seed_safety():
    rnd.seed(9)
    original = rnd.random((64))

    # Empty seed --------------
    rnd.seed(9)
    with ls():
        new = rnd.random((64))
        assert (original == new).all(), "Empty local seed changed outcome!"

    # Non-empty seed ----------
    rnd.seed(9)
    with ls(2):
        unimportant = rnd.random((64))

    new = rnd.random((64))
    assert (original == new).all(), "Non-empty local seed changed outcome!"

    # Nested seeds ------------
    with ls(9):
        with ls():
            new = rnd.random((64))
            assert (original == new).all(), "Nested empty seed changed outcome!"


if __name__ == "__main__":
    print("Testing rand_utils.py ...")
    test_seed_safety()
    print("All ok!")



"""Data uils.

This module contains functions to generate synthetic observations for the Argosim project.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>
          Samuel Gullin <gullin@ia.forth.gr>

"""

import jax.numpy as jnp
import numpy as np

from argosim.rand_utils import local_seed


def gauss_source(nx=512, ny=512, mu=np.array([0, 0]), sigma=np.eye(2), fwhm_pix=64):
    """Gauss source (JAX version).

    Function to generate a 2D Gaussian source.

    Parameters
    ----------
    nx : int
        The output image first dimension.
    ny : int
        The output image second dimension.
    mu : np.ndarray
        The mean of the Gaussian source in the [-1,1]x[-1,1] range.
    sigma : np.ndarray
        The covariance matrix of the Gaussian source.
    fwhm_pix : float
        The FWHM of the Gaussian source in pixels.

    Returns
    -------
    source : np.ndarray
        Image of size (nx,ny) containing the 2D Gaussian source.
    """
    fwhm = 2.355
    x, y = jnp.meshgrid(
        jnp.linspace(-fwhm / 2, fwhm / 2, nx), jnp.linspace(-fwhm / 2, fwhm / 2, ny)
    )

    sigma = sigma / (nx * ny) * fwhm_pix**2 / jnp.sqrt(jnp.linalg.det(sigma))

    X_unroll = jnp.array(
        [x.reshape(-1) - mu[0] * fwhm / 2, y.reshape(-1) - mu[1] * fwhm / 2]
    )
    sigminv = jnp.linalg.inv(sigma)
    sigminv.dot(X_unroll).shape
    Q = jnp.sum(jnp.multiply(X_unroll, sigminv.dot(X_unroll)), axis=0).reshape(ny, nx)
    return jnp.exp(-Q / 2)  # /(np.sqrt(2*np.pi*np.abs(np.linalg.det(sigma))))


def sigma2d(min_var=5, cov_lim=0.5, seed=None):
    """Sigma 2D.

    Function to generate a random 2D covariance matrix.

    Parameters
    ----------
    min_var : float
        The minimum variance of both gaussian components.
    cov_lim : float
        The limit of the covariance between gaussian components.
    seed : int
        Optional seed to set.

    Returns
    -------
    sigma : np.ndarray
        The 2D covariance matrix.
    """
    with local_seed(seed):
        var_1 = np.random.rand() + min_var
        # Limit eccentricity
        var_2 = np.random.rand() + min_var
        # Cov <= sqrt(var1 x var2)
        cov12 = (np.random.rand() * 2 - 1) * np.sqrt(var_1 * var_2) * cov_lim
    return np.array([[var_1, cov12], [cov12, var_2]])


def mu2d(seed=None):
    """Mu 2D.

    Function to generate a random 2D mean vector in the range [-1,1]x[-1,1].

    Parameters
    ----------
    seed : int
        Optional seed to set

    Returns
    -------
    mu : np.ndarray
        The 2D mean vector.
    """
    with local_seed(seed):
        mu = np.random.rand(2) * 2 - 1
    return mu


def random_source(shape, pix_size, seed=None):
    """Random source.

    Function to generate 2D Gaussian source with random mean and covariance.

    Parameters
    ----------
    shape : tuple
        The output image shape.
    pix_size : float
        The size in pixels of the Gaussian source.
    seed : int
        Optional seed to set

    Returns
    -------
    source : np.ndarray
        Image of size (nx,ny) containing the 2D Gaussian source.
    """
    with local_seed(seed):
        mu = mu2d()
        sigma = sigma2d()
    return gauss_source(shape[0], shape[1], mu, sigma, pix_size)


def n_source_sky(
    shape_px, fov, deg_size_list, source_intensity_list, seed=None, norm="none"
):
    """N source sky.

    Function to generate a sky image with multiple Gaussian sources at random positions.

    Parameters
    ----------
    shape_px : tuple
        The image size in pixels (Nx, Ny).
    fov: float
        the Image size in degrees. Is used with the first dimension of shape to
        calculate the source sizes.
    deg_size_list : list
        The size in degrees of the Gaussian sources.
    source_intensity_list : list
        The intensity of each Gaussian source in the final image.
        The sum of all the sources should be equal to 1 to have a normalized image.
    seed : int
        Optional seed to set
    norm : str
        The normalization method. Options are 'none', 'flux' and 'max'. Default is 'none'.

    Returns
    -------
    sky : np.ndarray
        Image of size (nx,ny) containing the sky model.
    """
    pix_per_deg = shape_px[0] / fov
    pix_size_list = [deg_size * pix_per_deg for deg_size in deg_size_list]
    with local_seed(seed):
        source_list = [
            random_source((shape_px[0], shape_px[1]), pix_size) * intensity
            for pix_size, intensity in zip(pix_size_list, source_intensity_list)
        ]

    sky = np.sum(source_list, axis=0)

    if norm == "flux":
        norm_sky = sky / np.sum(sky, axis=(0, 1))
    elif norm == "max":
        norm_sky = sky / np.max(sky)
    elif norm == "none":
        norm_sky = sky
    else:
        raise ValueError("Invalid normalization method. Use 'sum', 'max' or 'none'.")

    return norm_sky



"""Metrics utils.

This module contains utility functions to compute metrics between images.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import numpy as np
from skimage.metrics import structural_similarity as ssim_skimage


def mse(img1, img2):
    """Mean squared error.

    Function to compute the mean squared error between two images.

    Parameters
    ----------
    img1 : np.ndarray
        The first image.
    img2 : np.ndarray
        The second image.

    Returns
    -------
    mse : float
        The mean squared error between the two images.
    """
    return np.mean((img1 - img2) ** 2)


def residuals(img1, img2, absolute=True):
    """Residuals.

    Function to compute the residuals between two images.

    Parameters
    ----------
    img1 : np.ndarray
        The first image.
    img2 : np.ndarray
        The second image.

    Returns
    -------
    residual : np.ndarray
        The residuals between the two images.
    """
    res = img1 - img2
    return np.abs(res) if absolute else res


def ssim(img1, img2):
    """Structural similarity index.

    Function to compute the structural similarity index between two images.

    Parameters
    ----------
    img1 : np.ndarray
        The first image.
    img2 : np.ndarray
        The second image.

    Returns
    -------
    ssim : float
        The structural similarity index between the two images.
    """
    return ssim_skimage(np.array(img1), np.array(img2), data_range=np.max(img1) - np.min(img1))


def compute_metrics(img1, img2):
    """Compute metrics.

    Function to compute the metrics between two images.

    Parameters
    ----------
    img1 : np.ndarray
        The first image.
    img2 : np.ndarray
        The second image.

    Returns
    -------
    metrics : dict
        The metrics between the two images.
    """
    mse_val = mse(img1, img2)
    norm_sq = mse(img1, 0.0)
    metrics = {
        "mse": mse_val,
        "rel_mse": mse_val / norm_sq,
        "residual": residuals(img1, img2),
        "ssim": ssim(img1, img2),
    }
    return metrics


def fit_elliptical_beam(beam, threshold_ratio=0.5):
    """Fit elliptical beam.

    Fit an ellipse to the main lobe of the dirty beam using intensity thresholding.

    Parameters
    ----------
    beam : np.ndarray
        2D beam image.
    threshold_ratio : float
        Threshold fraction of the maximum intensity to define the bright region.

    Returns
    -------
    dict : dictionary
        Ellipse parameters: center, width, height, angle_deg, eccentricity.
    """
    max_val = np.max(np.abs(beam))
    threshold = threshold_ratio * max_val
    mask = np.abs(beam) >= threshold
    coords_yx = np.column_stack(np.where(mask))

    # Compute center and covariance matrix
    center_y, center_x = np.mean(coords_yx, axis=0)
    cov = np.cov(coords_yx, rowvar=False)

    # Eigendecomposition
    eigvals, eigvecs = np.linalg.eigh(cov)
    order = np.argsort(eigvals)[::-1]
    eigvals = eigvals[order]
    eigvecs = eigvecs[:, order]

    major_axis = eigvecs[:, 0]
    angle_rad = np.arctan2(major_axis[0], major_axis[1])
    angle_deg = np.degrees(angle_rad)

    # Axes
    width = 2 * np.sqrt(eigvals[0])
    height = 2 * np.sqrt(eigvals[1])

    # Eccentricity
    a, b = max(width, height), min(width, height)
    eccentricity = np.sqrt(1 - (b / a) ** 2)

    return {
        "center": (center_x, center_y),
        "width": width,
        "height": height,
        "angle_deg": angle_deg,
        "eccentricity": eccentricity,
    }


def mask_main_lobe_elliptical(beam, fit_result, scale=3.0):
    """Mask main lobe elliptical.

    Apply an elliptical mask to suppress the main lobe from a beam image.

    Parameters
    ----------
    beam : np.ndarray
        2D beam image.
    fit_result : dict
        Dictionary containing the ellipse parameters (center, width, height, angle_deg, eccentricity).
    scale : float
        Scale factor to enlarge or shrink the elliptical mask (default is 3).

    Returns
    -------
    beam_masked : np.ndarray
        Beam image with the main lobe masked (set to 0).
    """
    center_x, center_y = fit_result["center"]
    width = fit_result["width"] * scale
    height = fit_result["height"] * scale
    angle = np.radians(fit_result["angle_deg"])

    y, x = np.indices(beam.shape)
    x_shifted = x - center_x
    y_shifted = y - center_y

    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    x_rot = cos_angle * x_shifted + sin_angle * y_shifted
    y_rot = -sin_angle * x_shifted + cos_angle * y_shifted

    mask = (x_rot**2 / (width / 2) ** 2 + y_rot**2 / (height / 2) ** 2) < 1

    beam_masked = beam.copy()
    beam_masked[mask] = 0.0
    return beam_masked


def compute_sll(beam, fit_result=None, scale=3.0):
    """Compute sll.

    Compute the Side-lobe level (SLL) of a beam using an elliptical mask.

    Parameters
    ----------
    beam : np.ndarray
        The beam image (2D).
    fit_result : dict
        Dictionary containing the ellipse parameters (center, width, height, angle_deg, eccentricity). If None, it is computed from the beam.
    scale : float
        Scale factor for the elliptical mask.

    Returns
    -------
    sll_db : float
        Side-lobe level in dB.
    """
    if fit_result is None:
        fit_result = fit_elliptical_beam(beam)
    main_lobe_peak = np.max(np.abs(beam))
    beam_masked = mask_main_lobe_elliptical(np.abs(beam), fit_result, scale=scale)
    side_lobe_peak = np.max(beam_masked)
    sll_db = 10 * np.log10(side_lobe_peak / main_lobe_peak + 1e-12)
    return sll_db


def compute_fwhm(beam, fit_result=None):
    """Compute fwhm.

    Compute FWHM from beam.

    Parameters
    ----------
    beam : np.ndarray
        The beam image (2D).
    fit_result : dict
        Dictionary containing the ellipse parameters (center, width, height, angle_deg, eccentricity). If None, it is computed from the beam.

    Returns
    -------
    tuple
        FWHM along the semi-major and semi-minor axis of the beam.
    """
    if fit_result is None:
        fit_result = fit_elliptical_beam(beam)
    return fit_result["width"], fit_result["height"]


def compute_eccentricity(beam, fit_result=None):
    """Compute eccentricity.

    Compute the eccentricity of the beam main lobe.

    Parameters
    ----------
    beam : np.ndarray
        The beam image (2D).
    fit_result : dict
        Dictionary containing the ellipse parameters (center, width, height, angle_deg, eccentricity). If None, it is computed from the beam.

    Returns
    -------
    float
        Eccentricity of the beam. Ranges from 0 to 1, where 0 indicates a perfectly circular beam and 1 a completely degenerated beam.
    """
    if fit_result is None:
        fit_result = fit_elliptical_beam(beam)
    fwhm_a, fwhm_b = compute_fwhm(beam, fit_result)
    return np.sqrt(1 - (fwhm_b / fwhm_a) ** 2)


def compute_beam_metrics(beam):
    """Compute beam metrics.

    Compute main beam metrics: SLL, FWHM, and eccentricity.

    Parameters
    ----------
    beam : np.ndarray
        The beam image (2D).

    Returns
    -------
    dict
        Dictionary containing SLL (dB), FWHM (x, y), and eccentricity.
    """
    fit_result = fit_elliptical_beam(beam)
    sll = compute_sll(beam, fit_result)
    fwhm_a, fwhm_b = compute_fwhm(beam, fit_result)
    eccentricity = compute_eccentricity(beam, fit_result)
    return {
        "sll_db": sll,
        "fwhm": (fwhm_a, fwhm_b),
        "eccentricity": eccentricity,
    }



import numpy as np
import numpy.testing as npt

import argosim.metrics_utils as amu
from argosim.plot_utils import *


class TestMetricsUtils:

    obs_path = "src/argosim/tests/data/obs_sim_single_band.npy"
    sky_path = "src/argosim/tests/data/sky_model_exp.npy"

    mse_expect_same = 0.0
    mse_expect = 0.00010423388742998882
    mse_decimal = 10

    residual_expect_same = np.zeros_like(np.load(obs_path))
    residual_expect_path = "src/argosim/tests/data/residuals_obs_sky.npy"
    residual_decimal = 10

    rel_mse_expect_same = 0.0
    rel_mse_expect = 1.9389156923644497
    rel_mse_decimal = 10

    beam_path = "src/argosim/tests/data/dirty_beam_sim_single_band.npy"
    fit_beam_expect = {
        "center": (128.0, 128.0),
        "width": 8.542189128808278,
        "height": 2.774467796593848,
        "angle_deg": -14.678876771395638,
        "eccentricity": 0.9457841398924928,
    }
    fit_beam_decimal = 10
    fit_beam_decimal_center = 0

    beam_mask_path = "src/argosim/tests/data/dirty_beam_masked.npy"
    beam_mask_decimal = 10

    beam_metrics_expect = {
        "sll_db": -4.127083541527568,
        "fwhm": (8.542189128808278, 2.774467796593848),
        "eccentricity": 0.9457841398924928,
    }
    beam_metrics_decimal = 10

    def test_mse(self):
        obs = np.load(self.obs_path)
        sky = np.load(self.sky_path)
        mse_out = amu.mse(obs, sky)
        npt.assert_almost_equal(
            mse_out,
            self.mse_expect,
            decimal=self.mse_decimal,
            err_msg="MSE does not match expected value.",
        )

        mse_out_same = amu.mse(obs, obs)
        npt.assert_almost_equal(
            mse_out_same,
            self.mse_expect_same,
            decimal=self.mse_decimal,
            err_msg="MSE of identical images should be zero.",
        )

    def test_residuals(self):
        obs = np.load(self.obs_path)
        sky = np.load(self.sky_path)
        residual_expect = np.load(self.residual_expect_path)
        residuals_out = amu.residuals(obs, sky)
        npt.assert_almost_equal(
            residuals_out,
            residual_expect,
            decimal=self.residual_decimal,
            err_msg="Residuals do not match expected value.",
        )

        residuals_out_same = amu.residuals(obs, obs)
        npt.assert_almost_equal(
            residuals_out_same,
            self.residual_expect_same,
            decimal=self.residual_decimal,
            err_msg="Residuals of identical images should be zero.",
        )

    def test_compute_metrics(self):
        obs = np.load(self.obs_path)
        sky = np.load(self.sky_path)
        residual_expect = np.load(self.residual_expect_path)
        compute_metrics_out = amu.compute_metrics(obs, sky)
        compute_metrics_out_same = amu.compute_metrics(obs, obs)

        npt.assert_almost_equal(
            compute_metrics_out["mse"],
            self.mse_expect,
            decimal=self.mse_decimal,
            err_msg="MSE in compute_metrics does not match expected value.",
        )

        npt.assert_almost_equal(
            compute_metrics_out["rel_mse"],
            self.rel_mse_expect,
            decimal=self.rel_mse_decimal,
            err_msg="Relative MSE in compute_metrics does not match expected value.",
        )

        npt.assert_almost_equal(
            compute_metrics_out_same["rel_mse"],
            self.rel_mse_expect_same,
            decimal=self.rel_mse_decimal,
            err_msg="Relative MSE of identical images should be zero.",
        )

        npt.assert_almost_equal(
            compute_metrics_out["residual"],
            residual_expect,
            decimal=self.residual_decimal,
            err_msg="Residuals in compute_metrics does not match expected value.",
        )

    def test_fit_elliptical_beam(self):
        beam = np.load(self.beam_path)
        fit_beam_out = amu.fit_elliptical_beam(beam)

        for key, dec in [
            ("center", self.fit_beam_decimal_center),
            ("width", self.fit_beam_decimal),
            ("height", self.fit_beam_decimal),
            ("angle_deg", self.fit_beam_decimal),
            ("eccentricity", self.fit_beam_decimal),
        ]:
            npt.assert_almost_equal(
                fit_beam_out[key],
                self.fit_beam_expect[key],
                decimal=dec,
                err_msg=f"{key} of fitted beam does not match expected value.",
            )

    def test_mask_main_lobe_elliptical(self):
        beam_mask_expect = np.load(self.beam_mask_path)
        beam = np.load(self.beam_path)
        fit_beam_out = amu.fit_elliptical_beam(beam)
        beam_mask_out = amu.mask_main_lobe_elliptical(beam, fit_beam_out)
        npt.assert_almost_equal(
            beam_mask_out,
            beam_mask_expect,
            decimal=self.beam_mask_decimal,
            err_msg="Masked beam does not match expected value.",
        )

    def test_compute_sll(self):
        beam = np.load(self.beam_path)
        beam_sll_out = amu.compute_sll(beam)
        npt.assert_almost_equal(
            beam_sll_out,
            self.beam_metrics_expect["sll_db"],
            decimal=self.beam_metrics_decimal,
            err_msg="SLL of fitted beam does not match expected value.",
        )

    def test_compute_fwhm(self):
        beam = np.load(self.beam_path)
        beam_fwhm_out = amu.compute_fwhm(beam)
        npt.assert_almost_equal(
            beam_fwhm_out,
            self.beam_metrics_expect["fwhm"],
            decimal=self.beam_metrics_decimal,
            err_msg="FWHM of fitted beam does not match expected value.",
        )

    def test_compute_eccentricity(self):
        beam = np.load(self.beam_path)
        beam_ecc_out = amu.compute_eccentricity(beam)
        npt.assert_almost_equal(
            beam_ecc_out,
            self.beam_metrics_expect["eccentricity"],
            decimal=self.beam_metrics_decimal,
            err_msg="Eccentricity of fitted beam does not match expected value.",
        )

    def test_compute_beam_metrics(self):
        beam = np.load(self.beam_path)
        beam_metrics_out = amu.compute_beam_metrics(beam)

        for key in ["sll_db", "fwhm", "eccentricity"]:
            npt.assert_almost_equal(
                beam_metrics_out[key],
                self.beam_metrics_expect[key],
                decimal=self.beam_metrics_decimal,
                err_msg=f"{key} of fitted beam does not match expected value.",
            )



"""Imaging utils.

This module contains functions to perform radio interferometric imaging.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>
          Samuel Gullin <gullin@ia.forth.gr>

"""

import jax.numpy as jnp
import numpy as np
import numpy.random as rnd

from argosim.rand_utils import local_seed


def sky2uv(sky):
    """Sky to uv plane (JAX version).

    Function to compute the Fourier transform of the sky.

    Parameters
    ----------
    sky : np.ndarray
        The sky image.

    Returns
    -------
    sky_uv : np.ndarray
        The Fourier transform of the sky.
    """
    # return np.fft.fft2(sky)
    return jnp.fft.fftshift(jnp.fft.fft2(jnp.fft.ifftshift(sky)))


def scale_uv_samples(uv_samples, sky_uv_shape, fov_size):
    """Scale uv samples (JAX version).

    Function to scale the uv samples to pixel coordinates.

    Parameters
    ----------
    uv_samples : np.ndarray
        The uv samples coordinates in meters.
    sky_uv_shape : tuple
        The shape of the sky model in pixels.
    fov_size : tuple
        The field of view size in degrees.

    Returns
    -------
    uv_samples_indices : np.ndarray
        The indices of the uv samples in pixel coordinates.
    """
    max_u = (180 / jnp.pi) * sky_uv_shape[0] / (2 * fov_size[0])
    max_v = (180 / jnp.pi) * sky_uv_shape[1] / (2 * fov_size[1])
    uv_samples_indices = (
        jnp.rint(
            uv_samples[:, :2] / jnp.array([max_u, max_v]) / 2 * jnp.array(sky_uv_shape)
        )
        + jnp.array(sky_uv_shape) // 2
    )
    return uv_samples_indices


def check_uv_samples_range(uv_samples_indices, uv_samples, sky_uv_shape, fov_size):
    """Check uv samples range (JAX version).

    Function to check if the uv samples are within the uv-plane range.

    Parameters
    ----------
    uv_samples_indices : np.ndarray
        The indices of the uv samples in pixel coordinates.
    sky_uv_shape : tuple
        The shape of the sky model in pixels.
    uv_samples : np.ndarray
        The uv samples coordinates in meters.
    fov_size : tuple
        The field of view size in degrees.
    """
    sky_uv_shape_array = jnp.array(sky_uv_shape)
    if jnp.any(sky_uv_shape_array <= jnp.max(uv_samples_indices, axis=0)):
        max_uv = jnp.max(jnp.abs(uv_samples[:, :2]), axis=0)
        required_npix = jnp.ceil(max_uv * 2 * jnp.pi * jnp.array(fov_size) / 180)
        raise ValueError(
            f"uv samples lie out of the uv-plane. Required Npix > {required_npix}"
        )


def grid_uv_samples(
    uv_samples, sky_uv_shape, fov_size, mask_type="binary", weights=None
):
    """Grid uv samples (JAX version).

    Compute the uv sampling mask from the uv samples.

    Parameters
    ----------
    uv_samples : np.ndarray
        The uv samples coordinates in meters.
    sky_uv_shape : tuple
        The shape of the sky model in pixels.
    fov_size : tuple
        The field of view size in degrees.
    mask_type : str
        The type of mask to use. Choose between 'binary', 'histogram' and 'weighted'.
    weights : np.ndarray
        The weights to use for the mask type 'weighted'.

    Returns
    -------
    uv_mask : np.ndarray
        The uv sampling mask.
    uv_samples_indices : np.ndarray
        The indices of the uv samples in pixel coordinates.
    """
    uv_samples_indices = scale_uv_samples(uv_samples, sky_uv_shape, fov_size)
    # Check if the uv samples are within the uv-plane range
    check_uv_samples_range(uv_samples_indices, uv_samples, sky_uv_shape, fov_size)

    uv_mask = jnp.zeros(sky_uv_shape, dtype=jnp.complex128)

    # Convert uv_samples_indices to integer indices
    indices = jnp.array(uv_samples_indices, dtype=jnp.int32)

    if mask_type == "binary":
        uv_mask = uv_mask.at[indices[:, 1], indices[:, 0]].set(1 + 0j)
    elif mask_type == "histogram":
        uv_mask = uv_mask.at[indices[:, 1], indices[:, 0]].add(1 + 0j)
    elif mask_type == "weighted":
        assert weights is not None, "Weights must be provided for mask type 'weighted'."
        uv_mask = uv_mask.at[indices[:, 1], indices[:, 0]].add(
            weights[indices[:, 0], indices[:, 1]]
        )
    else:
        raise ValueError(
            "Invalid mask type. Choose between 'binary', 'histogram' and 'weighted'."
        )

    return uv_mask, uv_samples_indices


def uv2sky(uv):
    """Uv to sky (JAX version).

    Function to compute the inverse Fourier transform of the uv plane.

    Parameters
    ----------
    uv : np.ndarray
        The image in the uv/Fourier domain.

    Returns
    -------
    sky : np.ndarray
        The image in the sky domain.
    """
    return jnp.fft.fftshift(jnp.fft.ifft2(jnp.fft.ifftshift(uv))).real


def compute_visibilities_grid(sky_uv, uv_mask):
    """Compute visibilities gridded.

    Function to compute the visibilities from the fourier sky and the uv sampling mask.

    Parameters
    ----------
    sky_uv : np.ndarray
        The sky model in Fourier/uv domain.
    uv_mask : np.ndarray
        The uv sampling mask.

    Returns
    -------
    visibilities : np.ndarray
        Gridded visibilities on the uv-plane.
    """
    return sky_uv * uv_mask + 0 + 0.0j


def add_noise_uv(vis, uv_mask, sigma=0.1, seed=None):
    """Add noise in uv-plane.

    Function to add white gaussian noise to the visibilities in the uv-plane.

    Parameters
    ----------
    vis : np.ndarray
        The visibilities.
    mask : np.ndarray
        The uv sampling mask.
    sigma : float
        The standard deviation of the noise.
    seed : int
        Optional seed to set.

    Returns
    -------
    vis : np.ndarray
        The visibilities with added noise.
    """
    if sigma == 0.0:
        return vis

    with local_seed(seed):
        noise_sky = rnd.normal(0, sigma, vis.shape)
    noise_uv = sky2uv(noise_sky)

    return vis + compute_visibilities_grid(noise_uv, uv_mask)


def simulate_dirty_observation(
    sky, track, fov_size, multi_band=False, freqs=None, beam=None, sigma=0.2, seed=None
):
    """Simulate dirty observation.

    Function to simulate a radio observation of the sky model from the track uv-samples.

    Parameters
    ----------
    sky : np.ndarray
        The sky model image.
    track : np.ndarray
        The uv sampling points.
    fov_size : float
        The field of view size in degrees.
    multi_band : bool
        If True, simulate a multi-band observation.
    freqs : list
        The frequency list for the multi-band simulation.
    beam : Beam
        The beam object to apply to the sky, only used in multi-band simulations.
    sigma : float
        The standard deviation of the noise.
    seed : int
        Optional seed to set for reproducibility in noise realisation.

    Returns
    -------
    obs : np.ndarray
        The dirty observation(s).
    dirty_beam : np.ndarray
        The dirty beam(s).
    """
    if multi_band:
        assert freqs is not None, "Frequency list is required for multiband simulation"
        obs_multiband = []
        beam_multiband = []
        # Iterate over the frequency bands
        for f_, track_f in zip(freqs, track):
            # Apply beam to the sky
            if beam is not None:
                beam.set_fov(fov_size)
                beam.set_f(f_ / 1e9)
                beam_amplitude = beam.get_beam()
                sky_obs = sky * beam_amplitude
            else:
                sky_obs = sky
            # Transform to uv domain
            sky_uv = sky2uv(sky_obs)
            # Compute visibilities
            uv_mask, _ = grid_uv_samples(track_f, sky_uv.shape, (fov_size, fov_size))
            vis_f = compute_visibilities_grid(sky_uv, uv_mask)
            # Add noise
            vis_f = add_noise_uv(vis_f, uv_mask, sigma, seed=seed)
            # Back to image domain
            obs = uv2sky(vis_f)
            dirty_beam = uv2sky(uv_mask)

            obs_multiband.append(obs)
            beam_multiband.append(dirty_beam)

        obs = np.array(obs_multiband)
        dirty_beam = np.array(beam_multiband)
    else:
        sky_uv = sky2uv(sky)
        uv_mask, _ = grid_uv_samples(track, sky_uv.shape, (fov_size, fov_size))
        vis = compute_visibilities_grid(sky_uv, uv_mask)
        vis = add_noise_uv(vis, uv_mask, sigma, seed=seed)
        obs = uv2sky(vis)
        dirty_beam = uv2sky(uv_mask)

    return obs, dirty_beam



"""Beam utils.

This module contains the beam class to generate the primary beam of the antennas.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import matplotlib.pyplot as plt
import numpy as np


class CosCubeBeam:
    """Cosine cubed beam.

    Class to model the primary beam of the antennas using a cosine cubed function.

    Attributes
    ----------
    c : float
        The multiplicative constant of the cosine argument.
    f : float
        The frequency in GHz at which the beam is evaluated.
    grid_size : int
        The number of pixels of the simulated field of view.
    fov_deg : float
        The simulated field of view in degrees.

    """

    def __init__(self, c=0.2, f=1.0, n_pix=100, fov_deg=1.0):
        """Initialize the cosine cubed beam.

        Initialize the cosine cubed beam with the given parameters and check if the desired FOV is in agreement with the beam size.

        Parameters
        ----------
        c : float
            The multiplicative constant of the cosine argument.
        f : float
            The frequency in GHz at which the beam is evaluated.
        n_pix : int
            The number of pixels of the simulated field of view.
        fov_deg : float
            The simulated field of view in degrees.

        """
        self.c = c
        self.f = f
        self.grid_size = n_pix
        self.fov_deg = fov_deg

        # Check if the desired FOV is too large for the beam size
        self.check_fov()

    def __call__(self, l, m):
        """Cosine cubed beam.

        Function to compute the cosine cubed beam. The cosine cubed beam is modeled as:

        z = cos(Cf * sqrt(l^2 + m^2))^3,

        where Cf = c*f + c^2.

        Parameters
        ----------
        l : np.ndarray
            The l coordinate of the uv-plane.
        m : np.ndarray
            The m coordinate of the uv-plane.

        Returns
        -------
        z : np.ndarray
            The cosine cubed beam.

        """
        return np.cos(self.Cf() * np.sqrt(l**2 + m**2)) ** 3

    def check_fov(self):
        """Check FOV.

        Function to check if the desired FOV is too large for the beam size.

        """
        if np.sqrt(2) * self.fov_deg > self.beam_edge():
            print(
                "Warning: FOV diagonal lenght ({:.2f} deg) is larger than beam edge ({:.2f} deg).".format(
                    np.sqrt(2) * self.fov_deg, self.beam_edge()
                )
            )
            print(
                "Suggested FOV: {:.2f} deg".format(
                    np.floor(100 * self.beam_edge() / np.sqrt(2)) / 100
                )
            )

    def set_c(self, c):
        """Set c.

        Parameters
        ----------
        c : float
            The multiplicative constant of the cosine argument.

        """
        self.c = c
        self.check_fov()

    def set_f(self, f):
        """Set f.

        Parameters
        ----------
        f : float
            The frequency in GHz at which the beam is evaluated.

        """
        self.f = f
        self.check_fov()

    def set_fov(self, fov_deg):
        """Set FOV.

        Parameters
        ----------
        fov_deg : float
            The simulated field of view in degrees.

        """
        self.fov_deg = fov_deg
        self.check_fov()

    def Cf(self):
        """Cf.

        Function to compute the Cf parameter.

        Returns
        -------
        Cf : float
            The Cf parameter.

        """
        return self.c * self.f + self.c**2

    def get_mesh(self):
        """Get mesh.

        Function to compute the meshgrid of the lm image plane.

        Returns
        -------
        l : np.ndarray
            The l coordinate of the image plane.
        m : np.ndarray
            The m coordinate of the image plane.

        """
        l, m = np.meshgrid(
            np.linspace(-self.fov_deg, self.fov_deg, self.grid_size),
            np.linspace(-self.fov_deg, self.fov_deg, self.grid_size),
        )
        return l, m

    def get_beam(self):
        """Get beam.

        Function to compute the beam amplitude.

        Returns
        -------
        z : np.ndarray
            The beam amplitude.

        """
        l, m = self.get_mesh()
        z = self(l, m)
        return z

    def r_fov(self):
        """FOV radius.

        Function to compute the radius of the beam at -3 dB.

        Returns
        -------
        r_fov : float
            The radius of the beam at -3 dB.

        """
        return np.arccos((1 / np.sqrt(2)) ** (1 / 3)) / self.Cf()

    # def max_fov(self):
    #     """Max FOV.

    #     Compute the maximum FOF: radius at which the beam reaches 1/2 (max_FOV >> fov_3db).

    #     Returns
    #     -------
    #     max_fov : float
    #         The maximum FOV in degrees.

    #     """
    #     return np.arccos((.5)**(1/3))/self.Cf()

    def beam_edge(self):
        """Beam edge.

        Function to compute the beam edge (where the beam reaches zero).

        Returns
        -------
        beam_edge : float
            The beam edge in degrees.

        """
        return 0.5 * np.pi / self.Cf()

    def fov_solid_angle(self, r_fov=None):
        """FOV solid angle.

        Compute the solid angle of the FOV.

        Parameters
        ----------
        r_fov : float
            The beam width at -3 dB in degrees.

        Returns
        -------
        solid_angle : float
            The solid angle of the FOV in square degrees.

        """
        if r_fov is None:
            r_fov = self.r_fov()
        return np.pi * r_fov**2

    def plot_beam_1d(self, freqs):  # pragma: no cover
        """Plot beam 1D.

        Function to plot the simulated beam in 1D as a function of the angle.

        Parameters
        ----------
        freqs : list
            The list of frequencies in GHz at which the beam is evaluated.

        """
        f_old = self.f
        if freqs is None:
            freqs = [self.f]
        plt.figure()
        plt.hlines(
            1 / np.sqrt(2),
            -self.fov_deg,
            self.fov_deg,
            linestyle="--",
            color="k",
            alpha=0.5,
            label="-3dB",
        )
        cmap = plt.get_cmap("viridis")
        for i, f in enumerate(freqs):
            self.set_f(f)
            l = np.linspace(-self.fov_deg, self.fov_deg, self.grid_size)
            plt.plot(
                l, self(l, 0), color=cmap(i / len(freqs)), label="{} GHz".format(self.f)
            )
            plt.xlabel("Angle (deg)")
            plt.ylabel("Antenna Gain")
            plt.title("Cosine cubed beam")
            plt.legend()
            plt.grid()
        plt.show()
        self.set_f(f_old)

    def plot_beam_2d(self):  # pragma: no cover
        """Plot beam 2D.

        Function to plot the simulated beam in 2D over the field of view.

        """
        z = self.get_beam()
        fig, ax = plt.subplots()
        ax.imshow(
            z,
            extent=(-self.fov_deg, self.fov_deg, -self.fov_deg, self.fov_deg),
            vmin=0,
            vmax=1,
        )
        r_fov = self.r_fov()
        fov_contour = plt.Circle(
            (0, 0), r_fov, color="gray", alpha=0.7, linestyle="--", fill=False
        )
        fov_crop = plt.Rectangle(
            (-r_fov, -r_fov),
            2 * r_fov,
            2 * r_fov,
            color="k",
            alpha=0.5,
            linestyle="--",
            fill=False,
        )
        ax.add_artist(fov_contour)
        ax.add_artist(fov_crop)
        ax.set_xlabel("Angle (deg)")
        ax.set_ylabel("Angle (deg)")
        ax.set_title(
            r"Cosine cubed beam - FOV: {:.1f} $\deg^2$ @ {} GHz".format(
                self.fov_solid_angle(r_fov), self.f
            )
        )
        plt.show()

    def evaluate_beam_fit(self, freqs, norm_beam_list):  # pragma: no cover
        """Evaluate beam fit.

        Function to evaluate the beam fit at the given frequencies and compare it with the beam measurements.

        Parameters
        ----------
        freqs : list
            The list of frequencies in GHz at which the beam is evaluated.
        norm_beam_list : list
            The list of normalized beam measurements (amplitude not dB).

        """
        fig, ax = plt.subplots(1, 5, figsize=(4 * len(freqs), 4))
        cmap = plt.get_cmap("viridis")
        for i, norm_beam in enumerate(norm_beam_list):
            ax[i].plot(
                norm_beam[:, 0],
                norm_beam[:, 2],
                label="Main lobe",
                color=cmap(i / len(freqs)),
            )

            self.set_f(freqs[i])
            C_f = self.Cf()
            x = np.linspace(-1 / C_f, 1 / C_f, self.grid_size)

            ax[i].plot(x, self(0, x), color="k", linestyle="--", label="Fit")

            ax[i].hlines(
                1 / np.sqrt(2),
                -self.beam_edge(),
                self.beam_edge(),
                linestyle="--",
                color="gray",
                alpha=0.7,
                label="-3dB",
            )

            ax[i].set_ylim(0, 1)
            ax[i].set_xlim(-self.beam_edge(), self.beam_edge())
            ax[i].set_title(
                r"{:.1f} GHz - FOV: {:.1f} $\deg^2$".format(
                    self.f, self.fov_solid_angle()
                )
            )
            ax[i].legend()
            ax[i].set_xlabel("Angle (deg)")
            if i == 0:
                ax[i].set_ylabel("Gain (dB)")

        plt.show()
